#include "readout.h"
#include "gpio.h"
#include <future>
#include <string>
#include <iostream>
#include <iomanip>
#include <queue>
#include <cmath>
#include <omp.h>
#include <chrono>
#include <csignal>
#include <algorithm>

Readout::Readout(double max_ref_diff, unsigned interrupt_pin)
	: m_max_interval{max_ref_diff}
	, m_interrupt_pin{interrupt_pin}
	, gpio_thread{
			[&] {
			auto result{ setup() };
 			if (result != 0) {
				return result;
			}

			while (m_run) {
				result = read_tdc();
				if (result != 0) {
					break;
				}
			}
			return result;
		}
	}
	,analysis_thread{
			[&] {
			start_time = std::chrono::high_resolution_clock::now();
			while (m_run) {
			  //		  	  std::this_thread::sleep_for(process_loop_timeout);
			  dump_raw_tdc();
				//	process_queue();
			}
			if (!tdc_stop[1].empty() && !tdc_stop[0].empty()) {
			  dump_raw_tdc();
			  //process_queue(true);
			}
			end_time = std::chrono::high_resolution_clock::now();
			auto duration = std::chrono::duration_cast<std::chrono::microseconds>(end_time - start_time).count();
			std::cerr << evt_count << " events, ";
			std::cerr << duration << " ms, ";
			auto rate = static_cast<double>(evt_count) / static_cast<double>(duration) * 1e6;
			std::cerr << "rate is " << rate << "/s, " << rate * 60 << "/min, " << rate * 60 * 60 << "/h, " << rate * 60 * 60 * 24 << "/d, " << rate * 60 * 60 * 24 * 7 << "/w" << std::endl;
			std::cerr << "non processed events in queues: " << tdc_stop[0].size() << " " << tdc_stop[1].size() << " (should not be greate than " << max_queue_size << ")" << std::endl;
			return 0;
		}
	}{}

Readout::~Readout() {
	stop();
	gpio_thread.join();
	analysis_thread.join();
}

void Readout::stop()
{
	m_run = false;
}

auto Readout::setup()->int {
	if (gpx2) {
		std::cerr << "tried to create new gpx2 device but it is already created." << std::endl;
		return -1;
	}
	gpx2 = std::make_unique<SPI::GPX2_TDC::GPX2>();
	SPI::GPX2_TDC::Config conf{};
	conf.loadDefaultConfig();
	conf.BLOCKWISE_FIFO_READ = 1U;
	conf.COMMON_FIFO_READ = 0U;

	//conf.PIN_ENA_STOP3 = 0;
	//conf.PIN_ENA_STOP4 = 0;
	//conf.HIT_ENA_STOP3 = 0;
	//conf.HIT_ENA_STOP4 = 0;

	gpx2->init();

	int verbosity = 0;

	bool status = gpx2->write_config(conf);
	std::string config = gpx2->read_config();
	if (status && (config == conf.str())) {
		if (verbosity > 1) {
			std::cerr << "config written..." << std::endl;
		}
	}
	else {
		std::cerr << "tried to write:" << std::endl;
		print_hex(conf.str());
		std::cerr << "read back:" << std::endl;
		print_hex(config);
		std::cerr << "failed to write config!" << std::endl;
		return -1;
	}

	handler = std::make_unique<gpio>();
	gpio::setting pin_setting{};

	callback = handler->list_callback(pin_setting);

	handler->start();

	gpx2->init_reset();
	return 0;
}

auto Readout::read_tdc()->int {
	// checks if data is available on the gpx2 (by checking if interrupt pin is low)
	// then reads out data from the gpx2-tdc and puts it in the queue
	while (callback->read(m_interrupt_pin) != 0) {
		// while interrupt pin is high, do not readout (since there is no data available)
		// std::this_thread::sleep_for(std::chrono::microseconds(1));
	}
	for (unsigned i = 0; i < 4; i++) {//why was this only 4?
		auto now = std::chrono::system_clock::now();
		auto measurements = gpx2->read_results();
		for (unsigned j = 0; j < 4; j++) {
			if (measurements[j]) {
				measurements[j].ts = now;
				tdc_stop[j].emplace(std::move(measurements[j]));
			}
		}
	}
//	std::cout << std::endl;
	queue_condition.notify_all();

	return 0;
}

void Readout::dump_raw_tdc(){
  
    std::mutex mutex;
    std::unique_lock<std::mutex> lock {mutex};
  if (queue_condition.wait_for(lock,std::chrono::seconds(1))==std::cv_status::timeout)  {
    return;
  }

  std::array<std::queue<SPI::GPX2_TDC::Meas>,4> tdc_it(tdc_stop);
  std::cout << std::scientific << std::setprecision(5) ; //set output to 5 decimal places.
  
    std::cout << "\n ----- Full data set:"	<<std::endl;

  for (std::size_t j{0}; j<4; j++){
    while(!tdc_it[j].empty()){
      auto ref = static_cast<int32_t>(tdc_it[j].front().ref_index);
      auto fasttime = SPI::GPX2_TDC::inPicoseconds(tdc_it[j].front());
      tdc_it[j].pop();
      std::cout << j << " " << ref << " " << fasttime<< " " << std::endl;
    }
  }

  
  std::cout <<" -----  Matched Pairs:" <<std::endl;
  for (std::size_t i{0}; i<4; i=i+2){
    while(!tdc_stop[i].empty() && !tdc_stop[i+1].empty()){
      auto ref0 = static_cast<int32_t>(tdc_stop[i].front().ref_index);
      auto fasttime0 = SPI::GPX2_TDC::inPicoseconds(tdc_stop[i].front());
      auto ref1 = static_cast<int32_t>(tdc_stop[i+1].front().ref_index);
      auto fasttime1 = SPI::GPX2_TDC::inPicoseconds(tdc_stop[i+1].front());
      tdc_stop[i].pop();tdc_stop[i+1].pop();
      if (ref0-ref1==0) std::cout << i << " " << ref0 << " " << fasttime0<< " " << ref1 << " " << fasttime1<< " " << ref0-ref1 << " " << fasttime0-fasttime1<< std::endl;
    }
  }
    // auto result{ setup() };

  //remove remaining elements
  for (std::size_t i{0}; i<4; i++){
    while(!tdc_stop[i].empty()){
      tdc_stop[i].pop();
    }
  }
  gpx2->init_reset();
  evt_count += 8;
  return;
}
	
	
void Readout::process_queue(bool ignore_max_queue) {
	// checks ref_diff between ordered tdc_stop[0] and tdc_stop[1] channels.
	// Compares a few values from the queues, removes not matching timestamps and prints matching ones

	std::mutex mutex;
	std::unique_lock<std::mutex> lock {mutex};
	if (queue_condition.wait_for(lock,std::chrono::seconds(1))==std::cv_status::timeout || (tdc_stop[0].size() < max_queue_size && tdc_stop[1].size() < max_queue_size && !ignore_max_queue)) {
		return;
	}
	std::array<std::vector<SPI::GPX2_TDC::Meas>,4> data{};
	for (std::size_t i{0}; i<4; i++){
		while(tdc_stop[i].size()>min_queue_size || (ignore_max_queue && !tdc_stop[i].empty())){
			data[i].emplace_back(tdc_stop[i].front());
			tdc_stop[i].pop();
		}
	}
	// at this point one queue is full and we can check the content for 
	std::sort(data[0].begin(), data[0].end());
	std::unique(data[0].begin(), data[0].end());
	std::sort(data[1].begin(), data[1].end());
	std::unique(data[1].begin(), data[1].end());
	
	auto reset_ref = [&](){return static_cast<int32_t>(data[1].front().ref_index) - static_cast<int32_t>(data[0].front().ref_index);};
	while (!data[0].empty() && !data[1].empty()) {
		auto ref_diff = reset_ref();
		while (ref_diff > 1) {
			//std::cout << "ref_diff " << ref_diff << std::endl;
			data[0].erase(data[0].begin());
			//evt_count += 1U;
			if (data[0].empty()) {
				return;
			}
			ref_diff = reset_ref();
		}
		while (ref_diff < 1) {
			//std::cout << "ref_diff " << ref_diff << std::endl;
			data[1].erase(data[1].begin());
			//evt_count += 1U;
			if (data[1].empty()) {
				return;
			}
			ref_diff = reset_ref();
		}
		if (ref_diff == 0) {
			auto diff = SPI::GPX2_TDC::diff(data[0].front(), data[1].front());
			auto result0 = SPI::GPX2_TDC::inPicoseconds(data[0].front());
			auto result1 = SPI::GPX2_TDC::inPicoseconds(data[1].front());	
			if (fabs(diff) < m_max_interval) {
				evt_count += 1;
				std::cout << std::chrono::duration_cast<std::chrono::nanoseconds>(data[0].front().ts.time_since_epoch()).count();
//				std::cout << result0 << " ";
//                              std::cout << result1 << " ";
				std::cout << " ";
				std::cout << diff;
//				std::cout << " " << data[0].front().ref_index << " " << data[0].front().stop_result;
//				std::cout << " " << data[1].front().ref_index << " " << data[1].front().stop_result;
				std::cout << "\n";
			}
			data[0].erase(data[0].begin());
			data[1].erase(data[1].begin());
		}
	}
	gpx2->init_reset();

}
